package Tar;

use strict;
use Carp;
use diagnostics;

BEGIN {
    use Exporter ();
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

    $VERSION = 0.03;
    @ISA = qw(Exporter);
    @EXPORT = qw (&read_tar &write_tar);
    %EXPORT_TAGS = ();
    @EXPORT_OK = ();
}

=head1 NAME

Tar - module for manipulation of tar archives.

=head1 SYNOPSIS

  use Tar;

  $tar = Tar->new();
  $tar->add("file/foo.c");
  $tar->add("file/bar.c");
  $tar->write("files.tar");

=head1 DESCRIPTION

This module is definitely tentative, and several things will be
changed rather shortly. The exported routines will not be exported,
all the calls to croak() should be replaced with returning undef() and
putting error messages in a package global.

At the moment these methods are implemented:

=over 4

=item C<new()>

Returns a new Tar object. If given a filename as an argument, it will
try to load that as a tar file.

=item C<add(@filenamelist)>

Takes a list of filenames and adds them to the in-memory archive.

=item C<remove(@filenamelist)>

Removes any entries with names matching any of the given filenames
from the in-memory archive. String comparisons are done with C<eq>.

=item C<read('F<file.tar>')>

Try to read the given tarfile into memory. Will I<replace> any previous
content in C<$tar>!

=item C<write('F<file.tar>')>

Will write the in-memory archive to disk.

=item C<data()>

Returns the in-memory archive. This is a list of references to hashes,
the internals of which is not currently documented.

=back

=cut

use vars qw(@EXPORT_OK $tar_unpack_header $tar_pack_header $tar_header_length);

$tar_unpack_header 
  ='A100 A8 A8 A8 A12 A12 A8 A1 A100 A6 A2 A32 A32 A8 A8 A155';
$tar_pack_header 
  = 'a100a8a8a8a12a12a8ca100a6a2a32a32a8a8a155x12';
$tar_header_length = 512;

sub format_tar_entry;

###
### Non-method exported functions
###

sub read_tar {
    my ($filename) = @_;
    my @tarfile = ();
    my $i = 0;
    my $head;
    
    open(TAR, $filename) or croak "Can't open $filename: $!\n";
    
    while (read(TAR,$head,$tar_header_length)==$tar_header_length) {
	
	my ($name,		# string
	    $mode,		# octal number
	    $uid,		# octal number
	    $gid,		# octal number
	    $size,		# octal number
	    $mtime,		# octal number
	    $chksum,		# octal number
	    $typeflag,		# character
	    $linkname,		# string
	    $magic,		# string
	    $version,		# two bytes
	    $uname,		# string
	    $gname,		# string
	    $devmajor,		# octal number
	    $devminor,		# octal number
	    $prefix) = unpack($tar_unpack_header,$head);
		
	my ($data, $diff, $dummy);
	
	$mode = oct $mode;
	$uid = oct $uid;
	$gid = oct $gid;
	$size = oct $size;
	$mtime = oct $mtime;
	$chksum = oct $chksum;
	$devmajor = oct $devmajor;
	$devminor = oct $devminor;
	if (read(TAR,$data,$size)!=$size) {
	    croak "Read error on tarfile.\n";
	}
	$diff = $size%512;
	
	if ($diff!=0) {
	    read(TAR,$dummy,512-$diff); # Padding, throw away
	}

	next if $name eq ''; 
	
	$tarfile[$i++]={
			name => $name,		    
			mode => $mode,
			uid => $uid,
			gid => $gid,
			size => $size,
			mtime => $mtime,
			chksum => $chksum,
			typeflag => $typeflag,
			linkname => $linkname,
			magic => $magic,
			version => $version,
			uname => $uname,
			gname => $gname,
			devmajor => $devmajor,
			devminor => $devminor,
			prefix => $prefix,
			data => $data};
    }
    return @tarfile;
}

sub write_tar {
    my ($filename) = shift;
    my @tarfile = @_;
    my ($tmp);

    open(TAR, ">".$filename) or croak "Can't open $filename: $!\n";
    foreach (@tarfile) {
	$tmp = format_tar_entry $_;
	syswrite(TAR,$tmp,length($tmp));
    }
    close(TAR) or carp "Failed to close $filename, data may be lost: $!\n";
}

sub format_tar_entry {
    my ($ref) = shift;
    my ($tmp);
    
    $tmp = pack("a100a8a8a8a12a12a8a1a100",
		$ref->{name},
		sprintf("%6o ",$ref->{mode}),
		sprintf("%6o ",$ref->{uid}),
		sprintf("%6o ",$ref->{gid}),
		sprintf("%11o ",$ref->{size}),
		sprintf("%11o ",$ref->{mtime}),
		"        ",
		$ref->{typeflag},
		$ref->{linkname});
    $tmp .= pack("a6", $ref->{magic});
    $tmp .= '00';
    $tmp .= pack("a32",$ref->{uname});
    $tmp .= pack("a32",$ref->{gname});
    $tmp .= pack("a8",sprintf("%6o ",$ref->{devmajor}));
    $tmp .= pack("a8",sprintf("%6o ",$ref->{devminor}));
    $tmp .= pack("a155",$ref->{prefix});
    substr($tmp,148,6) = sprintf("%6o", unpack("%16C*",$tmp));
    substr($tmp,154,1) = "\0";
    $tmp .= "\0" x ($tar_header_length-length($tmp));
    $tmp .= $ref->{data};
    if ($ref->{size}>0) {
	$tmp .= "\0" x (512 - ($ref->{size}%512));
    }
    return $tmp;
}


###
### Methods
###

# Constructor. Reads tarfile if given an argument that's the name of a
# readable file.
sub new {
    my $class = shift;
    my $self = {};

    bless $self, $class;

    $self->{_filename} = undef;
    if (!defined $_[0]) {
	return $self;
    }
    if (-r $_[0]) {
	$self->{_data} = [read_tar $_[0]];
	$self->{_filename} = $_[0];
	return $self;
    }
    if (-e $_[0]) {
	carp "File exists but is not readable: $_[0]\n";
    }
    return $self;
}

# Return list with references to hashes representing the tar archive's
# component files.
sub data {
    my $self = shift;

    return @{$self->{_data}};
}

# Read a tarfile. Returns number of component files.
sub read {
    my $self = shift;
    my $file = $_[0];

    $self->{_filename} = undef;
    if (! -e $file) {
	carp "$file does not exist.\n";
	$self->{_data}=[];
	return undef;
    }
    elsif (! -r $file) {
	carp "$file is not readable.\n";
	$self->{_data}=[];
	return undef;
    }
    else {
	$self->{_data}=[read_tar $file];
	$self->{_filename} = $file;
	return scalar @{$self->{_data}};
    }
}

# Write a tar archive to file
sub write {
    my ($self) = shift @_;
    my ($file) = shift @_;

    unless ($file) {
	$file = $self->{_filename};
    }
    write_tar $file, @{$self->{_data}};
}

# Add files to the archive. Returns number of successfully added files.
sub add {
    my ($self) = shift;
    my (@files) = @_;
    my $file;
    my ($mode,$uid,$gid,$rdev,$size,$mtime,$data,$typeflag,$linkname);
    my $counter = 0;
    local ($/);

    undef $/;
    foreach $file (@files) {
	if ((undef,undef,$mode,undef,$uid,$gid,$rdev,$size,
	     undef,$mtime,undef,undef,undef) = stat($file)) {
	    $data = "";
	    $linkname = "";
	    if (-f $file) {	# Plain file
		$typeflag = 0;
		unless (open(FILE,$file)) {
		    next; # Can't open file, for some reason. Try next one.
		}
	    $data = <FILE>;
	    close FILE;
	    }
	    elsif (-d $file) {	# Directory
		$typeflag = 5;
	    }
	    elsif (-l $file) {	# Symlink
		$typeflag = 1;
		# This should be wrapped in an eval{} for portability
		$linkname = readlink $file;
	    }
	    elsif (-p $file) {	# Named pipe
		$typeflag = 6;
	    }
	    elsif (-S $file) {	# Socket
		$typeflag = 8;	# Bogus value, POSIX doesn't believe in sockets
	    }
	    elsif (-b $file) {	# Block special
		$typeflag = 4;
	    }
	    elsif (-c $file) {	# Character special
		$typeflag = 3;
	    }
	    else {		# Something else (like what?)
		$typeflag = 9;	# Also bogus value.
	    }
	    push(@{$self->{_data}},{
			       name => $file,		    
			       mode => $mode,
			       uid => $uid,
			       gid => $gid,
			       size => length $data,
			       mtime => $mtime,
			       chksum => "      ",
			       typeflag => $typeflag, 
			       linkname => $linkname,
			       magic => "ustar\0",
			       version => "00",
			       uname => (getpwuid($uid))[0],
			       gname => (getgrgid($gid))[0],
			       devmajor => 0, # We don't handle this yet
			       devminor => 0, # We don't handle this yet
			       prefix => "", # We don't handle this yet
			       'data' => $data,
			      });
	    	$counter++;	# Successfully added file
	}
	else {
	    next undef;	# stat failed
	}
    }
    return $counter;
}

sub remove {
    my ($self) = shift;
    my (@files) = @_;
    my $file;
    
    foreach $file (@files) {
	@{$self->{_data}} = grep {$_->{name} ne $file} @{$self->{_data}};
    }
    return $self;
}

1;
